/* This file is part of Cloudy and is copyright (C)1978-2007 by Gary J. Ferland
 * For conditions of distribution and use see copyright notice in license.txt */
/*HydroBranch compute branching ratios for Jason Ferguson's model hydrogen atom */
#include "cddefines.h"
#include "dense.h"
#include "phycon.h"
#include "path.h"
#include "hydrogenic.h"

t_hydrobranch::t_hydrobranch()
{
	const long VERSION_MAGIC = 20070109L;
	const static char chFile[] = "hydrobranch.dat";
	char chString[FILENAME_PATH_LENGTH_2];

	DEBUG_ENTRY( "t_hydrobranch()" );

	strcpy( chString, ( lgDataPathSet ? chDataPath : "" ) );
	strcat( chString, chFile );

	FILE *io;
	if( (io = fopen(chString,"r")) == NULL )
	{
		fprintf( ioQQQ, " Could not open %s for reading\n", chString );
		fprintf( ioQQQ, " Is the path set correctly?\n" );
		puts( "[Stop in t_hydrobranch]" );
		cdEXIT(EXIT_FAILURE);
	}

	long i=-1;

	bool lgErr = ( fscanf( io, "%ld", &i ) != 1 );
	if( lgErr || i != VERSION_MAGIC )
	{
		fprintf( ioQQQ, " File %s has incorrect version: %ld\n", chString, i );
		fprintf( ioQQQ, " I expected to find version: %ld\n", VERSION_MAGIC );
		puts( "[Stop in t_hydrobranch]" );
		cdEXIT(EXIT_FAILURE);
	}

	ReadBlock(io,Alow);
	ReadBlock(io,Blow);
	ReadBlock(io,Clow);
	ReadBlock(io,Ahigh);
	ReadBlock(io,Aint);
	ReadBlock(io,Bint);
	ReadBlock(io,Cint);
	ReadBlock(io,Azero);
	ReadBlock(io,Bzero);
	ReadBlock(io,Czero);

	/* initialize here so that lint does not complain about possibly
	 * uninitialized variables */
	int nd2=-1;

	lgErr = lgErr || ( fscanf( io, "%i", &nd2 ) != 1 );
	ASSERT( !lgErr && nd2 == NDIM2 );

	for( int j=0; j < nd2; j++ )
		lgErr = lgErr || ( fscanf( io, "%le", &DenCut[j] ) != 1 );

	ASSERT( !lgErr );

	DEBUG_EXIT( "t_hydrobranch()" );
}

void t_hydrobranch::ReadBlock(FILE* io,
			      double x[][NDIM2])
{
	int nd1, nd2;

	DEBUG_ENTRY( "ReadBlock()" );

	bool lgErr = ( fscanf( io, "%i %i", &nd1, &nd2 ) != 2 );

	ASSERT( !lgErr && nd1 == NDIM1 && nd2 == NDIM2 );

	for( int i=0; i < nd1; i++ )
		for( int j=0; j < nd2; j++ )
			lgErr = lgErr || ( fscanf( io, "%le", &x[i][j] ) != 1 );

	ASSERT( !lgErr );

	DEBUG_EXIT( "ReadBlock()" );
	return;
}

double t_hydrobranch::HydroBranchFunc(
	/* upper level quantum number, must be <= 15 */
	long int nHi, 
	/* lower quantum number, 2 or greater */
	long int nLo, 
	/* physical atomic number, H is 1 */
	long int ipZPhys) const
{
	/* function to evaluate branching ratios for hydrogen lines
	 * as a function of density
	 * Jason Ferguson, 1994-95,
	 * defined for lower 2 and above, and ipHi from 4 to 15 */

	long int nelem,
	  ipHi,
	  ipLo;
	double DenScale, 
	  RedMass, 
	  answer, 
	  den, 
	  denhigh, 
	  denint, 
	  denlow, 
	  denzero, 
	  t, 
	  teuse, 
	  tt, 
	  z7;

	DEBUG_ENTRY( "HydroBranchFunc()" );

	ASSERT( dense.eden > 0. );

	/* this allows quantum levels to be in any order and converts to C scale */
	ipLo = MIN2( nLo , nHi ) - 1;
	ipHi = MAX2( nLo , nHi ) - 1;

	/* only works for Balmer and higher lines since HS tables only had this */
	ASSERT(ipLo >= 1 );
	/* only works for Hbeta and higher */
	ASSERT( ipHi >= 3 );
	/* only tabulated up to 15 */
	ASSERT( ipHi <= 14);

	/* check on charge */
	ASSERT( ipZPhys > 0 );
	/* we will work on c scale for part of this */
	nelem = ipZPhys - 1;

	RedMass = dense.AtomicWeight[nelem]*1./
		(dense.AtomicWeight[nelem] + 1.);

	z7 = powi((double)ipZPhys,7);

	DenScale = log10(sqrt(2.*RedMass)/z7);

	den = log10(dense.eden) + DenScale;

	/* the fits only work down to den = log10( 0.001 ) - below that we are in ldl */
	den = MAX2( -2. , den );

	/* tt is used for the intermediate density fit
	 * at low densities the fit is good down to 100K
	 * at int. densities only 500K */
	t = phycon.te/POW2((double)ipZPhys);
	teuse = MAX2(100.,t);
	teuse = MIN2(3e4,teuse);
	t = log10(teuse);

	tt = phycon.te/POW2((double)ipZPhys);
	teuse = MAX2(500.,tt);
	teuse = MIN2(3e4,teuse);
	tt = log10(teuse);

	/* Compute the low and high density limits. */
	denlow = Alow[ipLo][ipHi] + t*(Blow[ipLo][ipHi] + t*Clow[ipLo][ipHi]);

	denhigh = Ahigh[ipLo][ipHi];

	/* Compute the really low den limit  */
	denzero = Azero[ipLo][ipHi] + t*(Bzero[ipLo][ipHi] + t*Czero[ipLo][ipHi]);

	/* the branching ratio at intermediate density: */
	denint = Aint[ipLo][ipHi] + tt*(Bint[ipLo][ipHi] + tt*Cint[ipLo][ipHi]);
	  

	if( den >= (DenCut[ipHi] + 2.) )
	{
		/* if the den>11.0 (for ipHi=4) then use just the high limit... */
		answer = denhigh;
	}
	else if( den >= DenCut[ipHi] )
	{
		/* between limits interpolate */
		answer = denint + (denhigh - denint)*(den - DenCut[ipHi])/2.0;
	}
	else if( den <= 4.0 )
	{
		/* if the den<4.0 then use the low limit... */
		if( ipZPhys != 1 && t <= 3.3 )
		{
			/* really low Te need an extra factor... */
			answer = denzero + (denlow - denzero)*(den - 2.0 - DenScale)/(2.0 - DenScale);
			if( answer <= 0. )
			{
				fprintf(ioQQQ," PROBLEM  hydrobranch low den limit ans %g %g %g %g %g\n",
					answer , denzero , denlow , den , DenScale);
				TotalInsanity();
			}
		}
		else
		{
			answer = denlow;
		}
	}
	else
	{
		/* else linerly interpolate between two limits... */
		answer = denlow + (denint - denlow)*(den - 4.0)/5.0;
	}

	if( answer <= 0. )
	{
		fprintf(ioQQQ,
			" HydroBranchFunc finds insane branching ratio, Z=%ld lo=%ld up=%ld val=%g\n" ,
			ipZPhys, ipLo+1, ipHi+1, answer);
		puts( "[Stop in HydroBranchFunc]" );
		cdEXIT(EXIT_FAILURE);
	}

	DEBUG_EXIT( "HydroBranchFunc()" );
	return answer;
}
