/* This file is part of Cloudy and is copyright (C)1978-2007 by Gary J. Ferland
 * For conditions of distribution and use see copyright notice in license.txt */
/**
 * \file hydrogenic.h                                                       
 * this file contains the variables for the model hydrogen ions,            
 * and prototypes for the series of routines that drive the atom            
 * the EXTERN structure hydro is defined here -                             
 * all H variables should migrate here  <BR>                                    
 *                                                                          
 * the hydrogenic lines, EmisLines[ipH_LIKE] are in taulines.h              
 *                                                                          */

/** the routine that calls the hydrogenic routines */
void Hydrogenic(void);

/**HydroLevel solve for ionization balance level populations of model hydrogen atom 
\param ipZ
*/
void HydroLevel(long int ipZ);

/**HydroLevelPop solve for departure coefs in hydrogen ionization balance 
\param nelem
*/
void HydroLevelPop(long int nelem );

/**HydroCollid evaluate collision rate for model hydrogen atom 
\param ipZ
*/
void HydroCollid(long int ipZ);

/**HydroT2Low called to do hydrogenic level populations when temp too low for matrix 
\param nelem
*/
void HydroT2Low(long int nelem );

/**HydroRecCool hydrogen recombination cooling 
\param n
\param ipZ
*/
double HydroRecCool(long int n, long int ipZ);

/** returns the ratio of recombination cooling to recombination coefficient 
 \param t the scaled temperature, T * n^2 / Z^2, n is prin quant number, Z is charge, 1 for H
*/
double HCoolRatio( 
	double t );

static const int NDIM1 = 14;
static const int NDIM2 = 15;

class t_hydrobranch : public Singleton<t_hydrobranch>
{
	friend class Singleton<t_hydrobranch>;
protected:
	t_hydrobranch();
private:
	/* Arrays A,B,C satisfy the equation y=A+B*x+C*x**2, x=LOG10(temp)
	 * for the low density branching ratio.
	 * Each line or pair of lines are for an upper level shown
	 * by the continue number on the left.  Reading from left to 
	 * right the fit for the ratio to 2, then 3, etc.. */
	double Alow[NDIM1][NDIM2];
	double Blow[NDIM1][NDIM2];
	double Clow[NDIM1][NDIM2];

	/* High density limit: */
	/* The array Ahigh is the high density branching ratio */
	double Ahigh[NDIM1][NDIM2];

	/* Intermediate limit: */
	/* Arrays Aint,Bint,Cint are the linear fits A + B*x + C*x^2, x=LOG10(Te)
	 * to the branching ratios at the density of DenCut below. */
	double Aint[NDIM1][NDIM2];
	double Bint[NDIM1][NDIM2];
	double Cint[NDIM1][NDIM2];

	/* Zero density limit: */
	/* The zero density BR for He+ fit to a1 + a2*x + a3*x^2, LOG10(Te) */
	double Azero[NDIM1][NDIM2];
	double Bzero[NDIM1][NDIM2];
	double Czero[NDIM1][NDIM2];

	/* the intermediate density for cutting the 
	 * interpolation.  Only used for ipHi=4,5,6,7,8. */
	double DenCut[NDIM2];

	void ReadBlock(FILE* in, double x[][NDIM2]);
public:
	double HydroBranchFunc(long int nHi, long int nLo, long int ipZPhys) const;
};

/**HydroBranch compute branching ratios for Jason's model hydrogen atom 
\param nHi     upper level quantum number, must be <= 15
\param nLo     lower quantum number, 2 or greater
\param ipZPhys physical atomic number, H is 1 
*/
inline double HydroBranch(long int nHi, long int nLo, long int ipZPhys)	
{
	return t_hydrobranch::Inst().HydroBranchFunc( nHi, nLo, ipZPhys );
}

/**HydroRecom photoionization, recombination, radiative rates for model hydrogen atom
\param ipZ
 */
void HydroRecom(long int ipZ);

/** HydroRenorm - renormalize H so that it agrees with the chemistry */
void HydroRenorm( void );

/** all of these are initialized in zero */
EXTERN struct t_hydro {

	/** flag to say (true) to produce emissivity for hydrogenic species,
	 * rather than intensity (the default).  Set with 
	 * hydrogen emissivity commmand.  default is false */
	bool lgHydEmiss;

	/** flag saying whether H atom is photo=false or collis ionized=true */
	bool lgHColionImp;

	/** lgHiPop2 flag set if H n=2 population gets large relative to ground
	 * pop2mx is maximum population of n=2 relative to ground */
	bool lgHiPop2;
	float pop2mx;

	/** Flag saying to reeval recom, because continuum lowering may otherwise leave
	 * us with zero'd levels. */
	bool lgReevalRecom;

	/** arrays for stark broadening in Puetter formalism */
	float **strkar/**strkar[LMHLVL+1][LMHLVL+1]*/;
	double **pestrk/**pestrk[LMHLVL+1][LMHLVL+1]*/;

	/** dstfe2lya is destruction probability for Lya onto FeII,
	 * net deexcitation of Lya but not ots destruction */
	float dstfe2lya;

	/** width of Lya */
	float HLineWidth;

	/** TexcLya is the excitation temperature of Lya */
	float TexcLya;

	/** nLyaHot is counts how ofter Lya hotter than gas */
	long int nLyaHot;

	/** TLyaMax is hottest */
	float TLyaMax, 
	/** TeLyaMax is electron temp at point where Lya max  */
	  TeLyaMax;

	/** nZTLaMax is the zone where this happened */
	long int nZTLaMax;

	/** chHTopType is the method.used to top off the H atom */
	char chHTopType[5];

	/** relative importance of photo ioniz from n=2 of H */
	float H_ion_frac_photo;

	/** largest fraction of ground state H destruction due to collisional ionization */
	float HCollIonMax;

	/** fraction of H ionizations due to ground collisions */
	float H_ion_frac_collis;

	/** array of lifetimes for HI levels, not including Lyman lines
	 * lifetime for principal quantum numbers
	 * [0] and [1] are set to -1e37, no defined,
	 * [2] is lya 
	 *
	 * this is used for full H iso sequence, so
	 * must be defined up through most number of levels
	 * needed for any atom on this sequence
	 */
	double HyLife[NHYDRO_MAX_LEVEL];

	/** cintot is total induced cooling over model */
	double cintot;

	/** lgHInducImp says whether or not induced recombination is important*/
	bool lgHInducImp;

	/** this is the D/H ratio, set with SET D/H command */
	double D2H_ratio;

	/** usually 1, set to 0 with hydrogen damping off command, scales rayleigh scat */
	float DampOnFac;

	/** remember induced fractions for hydrogen  */
	float FracInd;
	long int ndclev;
	float fbul;
	long int nbul;

	/** is continuum pumping of H lyman lines included?  yes, but turned off
	 * with atom h-like lyman pumping off command */
	bool lgLymanPumping;

	/** multiplicative scale factor for HI lyman line pump rate, takes into account
	 * possible emission lines - NB test against equal to 1.f in rt_lines_all to
	 * see if it has been set */
	float xLymanPumpingScaleFactor;

	} hydro;
